################################
##### Tests for Otter Logs #####
################################

import unittest

from otter.argparser import get_parser

from . import TestCase

parser = get_parser()

TEST_FILES_PATH = "test/test-logs/"

class TestLogs(TestCase):

    def test_something(self):
        pass



"""
Execution of an IPython notebook
"""

import os
import re
import ast
import copy

from unittest import mock
from contextlib import redirect_stdout, redirect_stderr
from IPython.display import display
from IPython.core.inputsplitter import IPythonInputSplitter

from .check_wrapper import CheckCallWrapper
from ..utils import hide_outputs

IGNORE_CELL_TAG = "otter_ignore"

def filter_ignored_cells(nb):
    """
    Filters out all cells in the notebook ``nb`` that are tagged with ``otter_ignore``. Returns a copy
    of the original notebook.

    Args:
        nb (``dict``): JSON representation of a notebook

    Returns:
        ``dict``: the notebook with cells removed
    """
    nb = copy.deepcopy(nb)

    for i, cell in enumerate(nb['cells']):
        metadata = cell.get("metadata", {})
        tags = metadata.get("tags", [])

        if IGNORE_CELL_TAG in tags:
            del nb['cells'][i]
    
    return nb

def execute_notebook(nb, secret='secret', initial_env=None, ignore_errors=False, cwd=None, test_dir=None, seed=None):
    """
    Executes a notebook and returns the global environment that results from execution

    Execute notebook & return the global environment that results from execution. If ``ignore_errors`` 
    is ``True``, exceptions are swallowed. ``secret`` contains random digits so ``check_results`` and 
    ``check`` are not easily modifiable. ``nb`` is passed in as a dictionary that's a parsed notebook

    Args:
        nb (``dict``): JSON representation of a notebook
        secret (``str``, optional): randomly generated integer used to rebind check function
        initial_env (``str``, optional): name of initial environment
        ignore_errors (``bool``, optional): whether exceptions should be ignored
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        seed (``int``, optional): random seed for intercell seeding
    
    Results:
        ``dict``: global environment resulting from executing all code of the input notebook
    """
    with hide_outputs():
        if initial_env:
            global_env = initial_env.copy()
        else:
            global_env = {}

        # add display from IPython
        global_env["display"] = display

        source = ""
        # if gradescope:
        #     source = "import sys\nsys.path.append(\"/autograder/submission\")\n"
        # el
        if cwd:
            source = f"import sys\nsys.path.append(r\"{cwd}\")\n"
            exec(source, global_env)
        if seed is not None:
            # source += "import numpy as np\nimport random\n"
            import numpy as np
            import random
            global_env["np"] = np
            global_env["random"] = random

        # Before rewriting AST, find cells of code that generate errors.
        # One round of execution is done beforehand to mimic the Jupyter notebook style of running
        # (e.g. code runs up to the point of execution).
        # The reason this is workaround is introduced is because once the
        # source code is parsed into an AST, there is no sense of local cells

        for cell in nb['cells']:
            if cell['cell_type'] == 'code':
                # transform the input to executable Python
                # FIXME: use appropriate IPython functions here
                isp = IPythonInputSplitter(line_input_checker=False)
                try:
                    code_lines = []
                    cell_source_lines = cell['source']
                    source_is_str_bool = False
                    if isinstance(cell_source_lines, str):
                        source_is_str_bool = True
                        cell_source_lines = cell_source_lines.split('\n')

                    for line in cell_source_lines:
                        # Filter out ipython magic commands
                        # Filter out interact widget
                        if not line.startswith('%'):
                            if "interact(" not in line and not re.search(r"otter\.Notebook\(.*?\)", line):
                                code_lines.append(line)
                                if source_is_str_bool:
                                    code_lines.append('\n')
                            elif re.search(r"otter\.Notebook\(.*?\)", line):
                                # TODO: move this check into CheckCallWrapper
                                # if gradescope:
                                #     line = re.sub(r"otter\.Notebook\(.*?\)", "otter.Notebook(\"/autograder/submission/tests\")", line)
                                # el
                                if test_dir:
                                    line = re.sub(r"otter\.Notebook\(.*?\)", f"otter.Notebook(\"{test_dir}\")", line)
                                else:
                                    line = re.sub(r"otter\.Notebook\(.*?\)", "otter.Notebook(\"/home/tests\")", line)
                                code_lines.append(line)
                                if source_is_str_bool:
                                    code_lines.append('\n')
                    if seed is not None:
                        cell_source = "np.random.seed({})\nrandom.seed({})\n".format(seed, seed) + isp.transform_cell(''.join(code_lines))
                    else:
                        cell_source = isp.transform_cell(''.join(code_lines))

                    # patch otter.Notebook.export so that we don't create PDFs in notebooks
                    # TODO: move this patch into CheckCallWrapper
                    m = mock.mock_open()
                    with mock.patch('otter.Notebook.export', m), mock.patch("otter.Notebook._log_event", m):
                        exec(cell_source, global_env)
                    source += cell_source
                except:
                    if not ignore_errors:
                        raise

        tree = ast.parse(source)
        # # CODE BELOW COMMENTED OUT BECAUSE the only check function is within the Notebook class
        # if find_check_assignment(tree) or find_check_definition(tree):
        #     # an empty global_env will fail all the tests
        #     return global_env

        # wrap check(..) calls into a check_results_X.append(check(..))
        transformer = CheckCallWrapper(secret)
        tree = transformer.visit(tree)
        ast.fix_missing_locations(tree)

        cleaned_source = compile(tree, filename="nb-ast", mode="exec")
        try:
            with open(os.devnull, 'w') as f, redirect_stdout(f), redirect_stderr(f):
                # patch otter.Notebook.export so that we don't create PDFs in notebooks
                m = mock.mock_open()
                with mock.patch('otter.Notebook.export', m), mock.patch("otter.Notebook._log_event", m):
                    exec(cleaned_source, global_env)
        except:
            if not ignore_errors:
                raise
        return global_env


import os
import unittest

class TestCase(unittest.TestCase):
    """
    Base class for Otter unit and integration tests that includes some helpful logging and custom 
    assert methods.
    """

    @classmethod
    def setUpClass(cls):
        print("\n" + ("=" * 70) + f"\nRunning Test Case {cls.__module__}.{cls.__name__}\n" + ("=" * 70))
        return super().setUpClass()

    def setUp(self):
        print("\n" + ("-" * 70) + f"\nRunning {self.id()}\n" + ("-" * 70))
        return super().setUp()

    def assertFilesEqual(self, p1, p2):
        try:
            with open(p1) as f1:
                with open(p2) as f2:
                    self.assertEqual(f1.read(), f2.read(), f"Contents of {p1} did not equal contents of {p2}")
        
        except UnicodeDecodeError:
            with open(p1, "rb") as f1:
                with open(p2, "rb") as f2:
                    self.assertEqual(f1.read(), f2.read(), f"Contents of {p1} did not equal contents of {p2}")

    def assertDirsEqual(self, dir1, dir2, ignore_ext=[]):
        self.assertTrue(os.path.exists(dir1), f"{dir1} does not exist")
        self.assertTrue(os.path.exists(dir2), f"{dir2} does not exist")
        self.assertTrue(os.path.isfile(dir1) == os.path.isfile(dir2), f"{dir1} and {dir2} have different type")

        if os.path.isfile(dir1):
            if os.path.splitext(dir1)[1] not in ignore_ext:
                self.assertFilesEqual(dir1, dir2)

        else:
            self.assertEqual(os.listdir(dir1), os.listdir(dir2), f"{dir1} and {dir2} have different contents")
            for f1, f2 in zip(os.listdir(dir1), os.listdir(dir2)):
                f1, f2 = os.path.join(dir1, f1), os.path.join(dir2, f2)
                self.assertDirsEqual(f1, f2, ignore_ext=ignore_ext)
    


"""
Docker container management for Otter Grade
"""

import subprocess
import re
import os
import shutil
import tempfile
import docker
import pandas as pd

from subprocess import PIPE
from concurrent.futures import ThreadPoolExecutor, wait

from .metadata import GradescopeParser
from .utils import simple_tar, get_container_file

def launch_parallel_containers(tests_dir, notebooks_dir, verbose=False, pdfs=None, reqs=None, 
    num_containers=None, image="ucbdsinfra/otter-grader", scripts=False, no_kill=False, output_path="./", 
    debug=False, seed=None, zips=False, meta_parser=None):
    """Grades notebooks in parallel docker containers

    This function runs ``num_containers`` docker containers in parallel to grade the student submissions
    in ``notebooks_dir`` using the tests in ``tests_dir``. It can additionally generate PDFs for the parts
    of the assignment needing manual grading. 

    Args:
        tests_dir (``str``): path to directory of tests
        notebooks_dir (``str``): path to directory of student submissions to be graded
        verbose (``bool``, optional): whether status messages should be printed to the command line
        pdfs (``str``, optional): string indicating what type of PDF to generate; defaults to ``None``
            indicating no PDFS but can be one of ``["unfiltered", "tags", "html"]`` corresponding to
            the type of cell filtering to perform upon generation
        reqs (``str``, optional): Path to requirements.txt file with non-standard packages needed
        num_containers (``int``, optional): The number of parallel containers that will be run
        image (``str``, optional): a docker image tag to be used for grading environment
        scripts (``bool``, optional): whether student submissions are Python scripts rather than
            Jupyter notebooks
        no_kill (``bool``, optional): whether the grading containers should be kept running after
            grading finishes
        output_path (``str``, optional): path at which to write grades CSVs copied from the container
        debug (``bool``, False): whether to run grading in debug mode (prints grading STDOUT and STDERR 
            from each container to the command line)
        seed (``int``, optional): a random seed to use for intercell seeding during grading
        zips (``bool``, False): whether the submissions are zip files formatted from ``Notebook.export``
        meta_parser (object, optional): a metadata parser instance; one of the classes defined in
            ``otter.metadata``

    Returns:
        ``list`` of ``pandas.core.frame.DataFrame``: the grades returned by each container spawned during
            grading
    """
    if not num_containers:
        num_containers = 4

    # list all notebooks in the dir
    dir_path = os.path.abspath(notebooks_dir)
    file_extension = (".zip", ".py", ".ipynb")[[zips, scripts, True].index(True)]
    notebooks = [os.path.join(dir_path, f) for f in os.listdir(dir_path) \
        if os.path.isfile(os.path.join(dir_path, f)) and f.endswith(file_extension)]

    # fix number of containers (overriding user input if necessary)
    if len(notebooks) < num_containers:
        num_containers = len(notebooks)

    # # calculate number of notebooks per container
    # num_per_group = int(len(notebooks) / num_containers)

    try:
        # create temp directories and add non-notebook files
        dirs = [tempfile.mkdtemp() for _ in range(num_containers)]

        # if GS export, get list of folder names to ignore when copying files
        if isinstance(meta_parser, GradescopeParser):
            ignore_folders = meta_parser.get_folders()
        else:
            ignore_folders = []

        # copy all non-notebook files into each temp directory
        for i in range(num_containers):
            for file in os.listdir(dir_path):
                fp = os.path.join(dir_path, file)
                if os.path.isfile(fp) and os.path.splitext(fp)[1] != file_extension:
                    shutil.copy(fp, dirs[i])
                elif os.path.isdir(fp) and fp not in ignore_folders:                    
                    shutil.copytree(fp, os.path.join(dirs[i], file))

        # copy notebooks in tmp directories
        for k, v in enumerate(notebooks):
            shutil.copy(v, dirs[k % num_containers])

        # execute containers in parallel
        pool = ThreadPoolExecutor(num_containers)
        futures = []
        for i in range(num_containers):
            futures += [pool.submit(grade_assignments, 
                tests_dir, 
                dirs[i], 
                str(i), 
                verbose=verbose, 
                pdfs=pdfs,
                # unfiltered_pdfs=unfiltered_pdfs, 
                # tag_filter=tag_filter,
                # html_filter=html_filter,
                reqs=reqs,
                image=image,
                scripts=scripts,
                no_kill=no_kill,
                output_path=output_path,
                debug=debug,
                seed=seed,
                zips=zips
            )]

        # stop execution while containers are running
        finished_futures = wait(futures)
    
    # cleanup temp directories on failure
    except:
        for i in range(num_containers):
            shutil.rmtree(dirs[i], ignore_errors=True)
        raise
    
    # cleanup temp directories
    else:
        for i in range(num_containers):
            shutil.rmtree(dirs[i])

    # return list of dataframes
    return [df.result() for df in finished_futures[0]]

def grade_assignments(tests_dir, notebooks_dir, id, image="ucbdsinfra/otter-grader", verbose=False, 
    pdfs=False, reqs=None, scripts=False, no_kill=False, output_path="./", debug=False, seed=None, zips=False):
    """
    Grades multiple assignments in a directory using a single docker container. If no PDF assignment is 
    wanted, set all three PDF params (``unfiltered_pdfs``, ``tag_filter``, and ``html_filter``) to ``False``.

    Args:
        tests_dir (``str``): path to directory of tests
        notebooks_dir (``str``): path to directory of student submissions to be graded
        id (``int``): grading container index
        image (``str``, optional): a docker image tag to be used for grading environment
        verbose (``bool``, optional): whether status messages should be printed to the command line
        pdfs (``str``, optional): string indicating what type of PDF to generate; defaults to ``None``
            indicating no PDFS but can be one of ``["unfiltered", "tags", "html"]`` corresponding to
            the type of cell filtering to perform upon generation
        reqs (``str``, optional): Path to requirements.txt file with non-standard packages needed
        scripts (``bool``, optional): whether student submissions are Python scripts rather than
            Jupyter notebooks
        no_kill (``bool``, optional): whether the grading containers should be kept running after
            grading finishes
        output_path (``str``, optional): path at which to write grades CSVs copied from the container
        debug (``bool``, False): whether to run grading in debug mode (prints grading STDOUT and STDERR 
            from each container to the command line)
        seed (``int``, optional): a random seed to use for intercell seeding during grading
        zips (``bool``, False): whether the submissions are zip files formatted from ``Notebook.export``

    Returns:
        ``pandas.core.frame.DataFrame``: A dataframe of file to grades information
    """
    client = docker.from_env()
    container = client.containers.run(image, detach=True, tty=True)

    try:        
        container_id = container.id[:12]

        if verbose:
            print(f"Launched container {container_id}...")

        with simple_tar(notebooks_dir) as tarf:
            container.put_archive("/home", tarf)
            container.exec_run(f"mv /home/{os.path.basename(notebooks_dir)} /home/notebooks")
        
        # copy the test files to the container
        if tests_dir is not None:
            with simple_tar(tests_dir) as tarf:
                container.put_archive("/home", tarf)
                container.exec_run(f"mv /home/{os.path.basename(tests_dir)} /home/tests")

        # copy the requirements file to the container
        if reqs:
            if verbose:
                print(f"Installing requirements in container {container_id}...")
            
            with simple_tar(reqs) as tarf:
                container.put_archive("/home", tarf)

            # install requirements
            exit_code, output = container.exec_run("pip3 install -r /home/requirements.txt")
            assert exit_code == 0, f"Container {container_id} failed to install requirements:\n{output.decode('utf-8')}"

            if debug:
                print(output.decode("utf-8"))

        if verbose:
            print(f"Grading {('notebooks', 'scripts')[scripts]} in container {container_id}...")
        
        # Now we have the notebooks in home/notebooks, we should tell the container to execute the grade command...
        grade_command = ["python3", "-m", "otter.grade", "/home/notebooks"]

        # if we want PDF output, add the necessary flag
        if pdfs:
            grade_command += ["--pdf", pdfs]
        
        # seed
        if seed is not None:
            grade_command += ["--seed", str(seed)]

        # if we are grading scripts, add the --script flag
        if scripts:
            grade_command += ["--scripts"]
        
        if debug:
            grade_command += ["--verbose"]

        if zips:
            grade_command += ["--zips"]

        exit_code, output = container.exec_run(grade_command)
        assert exit_code == 0, f"Container {container_id} failed with output:\n{output.decode('utf-8')}"

        if debug:
            print(output.decode("utf-8"))
        
        if verbose:
            print(f"Copying grades from container {container_id}...")

        # get the grades back from the container and read to date frame so we can merge later
        with get_container_file(container, "/home/notebooks/grades.csv") as f:
            df = pd.read_csv(f)

        if pdfs:
            pdf_folder = os.path.join(os.path.abspath(output_path), "submission_pdfs")
            os.makedirs(pdf_folder, exist_ok=True)
            
            # copy out manual submissions
            for pdf in df["manual"]:
                local_pdf_path = os.path.join(pdf_folder, os.path.basename(pdf))
                with get_container_file(container, pdf) as pdf_file, open(local_pdf_path, "wb+") as f:
                    f.write(pdf_file.read())

            df["manual"] = df["manual"].str.replace("/home/notebooks", os.path.basename(pdf_folder))

        if not no_kill:
            if verbose:
                print(f"Stopping container {container_id}...")

            # cleanup the docker container
            container.stop()
            container.remove()
        
        client.close()

    except:

        # delete the submission PDFs on failure
        if pdfs and "pdf_folder" in locals():
            shutil.rmtree(pdf_folder)

        if not no_kill:
            if verbose:
                print(f"Stopping container {container_id}...")

            # cleanup the docker container
            container.stop()
            container.remove()
        
        client.close()
        
        raise
    
    return df



"""
Database creation and setup for Otter Service
"""

import yaml
import csv
import psycopg2
import hashlib
import os

from psycopg2 import connect, extensions, sql
from psycopg2.errors import DuplicateTable

from .utils import connect_db

def create_users(filepath, host=None, username=None, password=None, conn=None):
    """
    Inserts usernames from ``filepath`` into Otter Service's database

    Args:
        filepath (``str``): Path to CSV files with usernames in first column, passwords in second
        host (``str``, optional): Hostname where Postgres is running
        username (``str``, optional): Username for writing to Postgres database
        password (``str``, optional): Password corresponding to provided username
    """
    with open(filepath, newline='') as csvfile:
        filereader = csv.reader(csvfile, delimiter=',', quotechar='|')
        # TODO: fill in the arguments below
        if conn is None:
            conn = connect_db(host, username, password)
        cursor = conn.cursor()
        for row in filereader:
            username, password = row[:2]
            password = hashlib.sha256(password.encode()).hexdigest()
            if username.lower() == "username":
                # skip heading
                continue
            insert_command = """
                INSERT INTO users (username, password) VALUES (\'{}\', \'{}\')
                ON CONFLICT (username)
                DO UPDATE SET password = \'{}\'
                """.format(username, password, password)
            cursor.execute(insert_command)

def remove_users(filepath, host=None, username=None, password=None, conn=None):
    """
    Removes users specified in file ``filepath``
    
    Args:
        filepath (``str``): Path to CSV files with usernames in first column, passwords in second
        host (``str``, optional): Hostname where Postgres is running
        username (``str``, optional): Username for writing to Postgres database
        password (``str``, optional): Password corresponding to provided username
    """
    with open(filepath, newline='') as csvfile:
        filereader = csv.reader(csvfile, delimiter=',', quotechar='|')
        # TODO: fill in the arguments below
        if conn is None:
            conn = connect_db(host, username, password)
        cursor = conn.cursor()
        remove_count = 0
        for row in filereader:
            username, password = row[:2]
            password = hashlib.sha256(password.encode()).hexdigest()
            if username.lower() == "username":
                # skip heading
                continue
            insert_command = """
                DELETE FROM users WHERE username = \'{}\'
                """.format(username)
            cursor.execute(insert_command)

def main(args, conn=None, close_conn=True):
    """
    Creates Otter Service database and tables

    Args:
        args (``argparse.Namespace``): parse command-line arguments
        conn (``psycopg2.connection``, optional): Postgres connection
        close_conn (``bool``, optional): whether to close the databse connection after running; 
            default ``True``
    """
    # if args.missing_packages:
    #     raise ImportError(
    #         "Missing some packages required for otter service. "
    #         "Please install all requirements at "
    #         "https://raw.githubusercontent.com/ucbds-infra/otter-grader/master/requirements.txt"
    #     )
    
    pgconn = connect_db(args.db_host, args.db_user, args.db_pass, args.db_port, db='postgres')

    try:
        pgconn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        cursor = pgconn.cursor()
        cursor.execute('CREATE DATABASE otter_db')
    except psycopg2.errors.DuplicateDatabase:
        pass
    finally: 
        cursor.close()
        pgconn.close()

    if conn is None:
        conn = connect_db(args.db_host, args.db_user, args.db_pass, args.db_port)

    # conn = connect_db(args.db_host, args.db_user, args.db_pass, args.db_port)

    conn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    cursor = conn.cursor()

    queries = [
        '''
        CREATE TABLE users (
            user_id SERIAL PRIMARY KEY,
            api_keys VARCHAR[] CHECK (cardinality(api_keys) > 0),
            username TEXT UNIQUE,
            password VARCHAR,
            email TEXT UNIQUE,
            CONSTRAINT has_username_or_email CHECK (username IS NOT NULL or email IS NOT NULL)
        )
        ''',
        '''
        CREATE TABLE classes (
            class_id TEXT PRIMARY KEY,
            class_name TEXT NOT NULL
        )
        ''',
        '''
        CREATE TABLE assignments (
            assignment_id TEXT NOT NULL,
            class_id TEXT REFERENCES classes (class_id) NOT NULL,
            assignment_name TEXT NOT NULL,
            seed INTEGER,
            PRIMARY KEY (assignment_id, class_id)
        )
        ''',
        '''
        CREATE TABLE submissions (
            submission_id SERIAL PRIMARY KEY,
            assignment_id TEXT NOT NULL,
            class_id TEXT NOT NULL,
            user_id INTEGER REFERENCES users(user_id) NOT NULL,
            file_path TEXT NOT NULL,
            timestamp TIMESTAMP NOT NULL,
            score JSONB,
            FOREIGN KEY (assignment_id, class_id) REFERENCES assignments (assignment_id, class_id)
        )
        '''
    ]

    for query in queries:
        try:
            cursor.execute(query)
        except DuplicateTable:
            pass
        
    cursor.close()
    if close_conn:
        conn.close()



"""
Docker image building for Otter Service
"""

import subprocess
import shutil
import os
import yaml

from subprocess import PIPE, DEVNULL
from io import BytesIO
from jinja2 import Template
from psycopg2.errors import UniqueViolation

from .utils import connect_db

DOCKERFILE_TEMPLATE = Template("""
FROM {{ image }}
RUN mkdir /home/notebooks
ADD {{ test_folder_path }} /home/tests{% if global_requirements %}
ADD {{ global_requirements }} /home
RUN pip3 install -r /home/{{ global_requirements_filename }}{% endif %}{% if requirements %}
ADD {{ requirements }} /home
RUN pip3 install -r /home/{{ requirements_filename }}{% endif %}{% for file in files %}
ADD {{ file }} /home/notebooks{% endfor %}
""")

def write_class_info(class_id, class_name, conn):
    """
    Writes the given ``class_name`` to the database

    Args:
        class_id (``str``): the class id for the database
        class_name (``str``): name of new class to add to the database
        conn (``psycopg2.connection``): connection object for database
    
    Returns:
        ``str``: class id for newly added class
    """
    cursor = conn.cursor()
    insert_command = "INSERT INTO classes (class_id, class_name) \
        VALUES(%s, %s)"#.format(class_name)
    try:
        cursor.execute(insert_command, (class_id, class_name))
    except UniqueViolation:
        pass
    conn.commit()
    cursor.close()
    return class_id

def write_assignment_info(assignment_id, class_id, assignment_name, seed, conn):
    """
    Inserts or updates an assignment with p-key ``(class_id, assignment_id)`` in the ``assignments`` 
    table of the database.

    Args:
        assignment_id (``str``): assignment id which will be created or updated
        class_id (``str``): class id of the assignment
        assignment_name (``str``): assignment name to insert/update
        conn (``psycopg2.connection``): connection object for database
    """
    cursor = conn.cursor()
    find_sql_command = """SELECT * FROM assignments
    WHERE assignment_id = '{}' AND class_id = '{}'
    """.format(assignment_id, class_id)
    cursor.execute(find_sql_command)

    # if seed is None:
    #     seed = "NULL"

    # If conflict on assignment_id, class_id
    # Update assignment_name
    if cursor.rowcount == 1:
        sql_command = """UPDATE assignments
                        SET assignment_name = %s, seed = %s
                        WHERE assignment_id = %s AND class_id = %s
                        """#.format(assignment_id, class_id, assignment_name, assignment_name)
        cursor.execute(sql_command, (assignment_name, seed, assignment_id, class_id))
    # Else, just insert
    else:
        sql_command = """INSERT INTO assignments (assignment_id, class_id, assignment_name, seed)
                        VALUES (%s, %s, %s, %s)
                        """#.format(assignment_id, class_id, assignment_name, assignment_name)
        cursor.execute(sql_command, (assignment_id, class_id, assignment_name, seed))

    conn.commit()
    cursor.close()

def main(args, conn=None, close_conn=True):
    """
    Inserts/updates assignment records in the database and builds Docker images from configuration
    repository based on a YAML configuation file.

    Args:
        args (``argparse.Namespace``): parsed command-line arguments
        conn (``psycopg2.connection``, optional): Postgres connection
        close_conn (``bool``, optional): whether to close the databse connection after running; 
            default ``True``
    """
    # if args.missing_packages:
    #     raise ImportError(
    #         "Missing some packages required for otter service. "
    #         "Please install all requirements at "
    #         "https://raw.githubusercontent.com/ucbds-infra/otter-grader/master/requirements.txt"
    #     )

    repo_path = args.repo_path
    assert os.path.exists(repo_path) and os.path.isdir(repo_path), "{} does not exist or is not a directory".format(repo_path)

    current_dir = os.getcwd()
    os.chdir(repo_path)

    # parse conf.yml
    assert os.path.isfile("conf.yml"), "conf.yml does not exist"
    with open("conf.yml") as f:
        config = yaml.safe_load(f)

    assignments = config["assignments"]
    assignment_cfs = [(a["name"], a["assignment_id"], a.get("seed", None)) for a in assignments]
    ids = [a["assignment_id"] for a in assignments]
    assert len(ids) == len(set(ids)), "Found non-unique assignment IDs in conf.yml"

    # Use one global connection for all db-related commands
    if conn is None:
        conn = connect_db(args.db_host, args.db_user, args.db_pass, args.db_port)
    class_id = write_class_info(config["class_id"], config["class_name"], conn)

    # write to the database
    for name, assignment_id, seed in assignment_cfs:
        write_assignment_info(assignment_id, class_id, name, seed, conn)

    # TODO: start building docker images
    for a in assignments:
        requirements = a["requirements"] if "requirements" in a else ""
        global_requirements = config["requirements"] if "requirements" in config else ""

        dockerfile = DOCKERFILE_TEMPLATE.render(
            image = args.image,
            test_folder_path = a["tests_path"],
            test_folder_name = os.path.split(a["tests_path"])[1],
            requirements = requirements,
            requirements_filename = os.path.split(requirements)[1],
            global_requirements = global_requirements,
            global_requirements_filename = os.path.split(global_requirements)[1],
            files = a.get("files", [])
        )

        print("Building Docker image {}".format(class_id + "-" + a["assignment_id"]))
        
        # Build the docker image
        echo_dockerfile = subprocess.Popen(["echo", dockerfile], stdout=PIPE)
        echo_dockerfile.wait()
        
        build_out = subprocess.Popen(
            ["docker", "build", "-f", "-", ".", "-t", class_id + "-" + a["assignment_id"]],
            stdin=echo_dockerfile.stdout, 
            stdout=DEVNULL if args.quiet else None
        )
        build_out.wait()
        
        echo_dockerfile.stdout.close()

        print("Built Docker image {}".format(class_id + "-" + a["assignment_id"]))
    
    if close_conn:
        conn.close()
    
    os.chdir(current_dir)



#################################
##### Tests for otter check #####
#################################

import os
import unittest
import subprocess
import contextlib

from textwrap import dedent
from subprocess import PIPE
from glob import glob
from io import StringIO
from unittest import mock

from otter import Notebook
from otter.argparser import get_parser
from otter.check import main as check

from . import TestCase

parser = get_parser()

TEST_FILES_PATH = "test/test-check/"

class TestCheck(TestCase):

    def test_otter_check_script(self):
        """
        Checks that the script checker works
        """
        # run for each individual test
        for file in glob(TEST_FILES_PATH + "tests/*.py"):
            check_command = ["check",
                TEST_FILES_PATH + "file0.py", 
                "-q", os.path.split(file)[1][:-3],
                "-t", os.path.split(file)[0]
            ]
            args = parser.parse_args(check_command)
            args.func = check

            # capture stdout
            output = StringIO()
            with contextlib.redirect_stdout(output):

                # mock block_print otherwise they interfere with capture of stdout
                with mock.patch("otter.check.block_print"):
                    args.func(args)
                    if os.path.split(file)[1] != "q2.py":
                        self.assertEqual(
                            output.getvalue().strip().split("\n")[-1].strip(), 
                            "All tests passed!", 
                            "Did not pass test at {}".format(file)
                        )

        # run checker command
        check_command = check_command = ["check",
            TEST_FILES_PATH + "file0.py", 
            "-t", TEST_FILES_PATH + "tests"
        ]
        args = parser.parse_args(check_command)
        args.func = check

        # capture stdout
        output = StringIO()
        with contextlib.redirect_stdout(output):

            # mock block_print otherwise they interfere with capture of stdout
            with mock.patch("otter.check.block_print"):
                args.func(args)
                self.assertEqual(
                    output.getvalue().strip(), 
                    dedent("""\
                        [0.         0.02002002 0.04004004 0.06006006 0.08008008]
                        4 of 5 tests passed

                        Tests passed:
                            q1 q3 q4 q5 


                        Tests failed: 
                           test/test-check/tests/q2.py

                        Test result:
                        Trying:
                            1 == 1
                        Expecting:
                            False
                        **********************************************************************
                        Line 2, in test/test-check/tests/q2.py 1
                        Failed example:
                            1 == 1
                        Expected:
                            False
                        Got:
                            True"""), 
                    "Did not pass correct tests"
                )


    def test_otter_check_notebook(self):
        """
        Checks that the script checker works
        """
        # run for each individual test
        for file in glob(TEST_FILES_PATH + "tests/*.py"):
            check_command = ["check",
                TEST_FILES_PATH + "test-nb.ipynb", 
                "-q", os.path.split(file)[1][:-3],
                "-t", os.path.split(file)[0]
            ]
            args = parser.parse_args(check_command)
            args.func = check

            # capture stdout
            output = StringIO()
            with contextlib.redirect_stdout(output):

                # mock block_print otherwise they interfere with capture of stdout
                with mock.patch("otter.check.block_print"):
                    args.func(args)
                    if os.path.split(file)[1] != "q2.py":
                        self.assertEqual(
                            output.getvalue().strip().split("\n")[-1].strip(), 
                            "All tests passed!", 
                            "Did not pass test at {}".format(file)
                        )

        # run checker command
        check_command = check_command = ["check",
            TEST_FILES_PATH + "file0.py", 
            "-t", TEST_FILES_PATH + "tests"
        ]
        args = parser.parse_args(check_command)
        args.func = check

        # capture stdout
        output = StringIO()
        with contextlib.redirect_stdout(output):

            # mock block_print otherwise they interfere with capture of stdout
            with mock.patch("otter.check.block_print"):
                args.func(args)
                self.assertEqual(
                    output.getvalue().strip(), 
                    dedent("""\
                        [0.         0.02002002 0.04004004 0.06006006 0.08008008]
                        4 of 5 tests passed

                        Tests passed:
                            q1 q3 q4 q5 


                        Tests failed: 
                           test/test-check/tests/q2.py

                        Test result:
                        Trying:
                            1 == 1
                        Expecting:
                            False
                        **********************************************************************
                        Line 2, in test/test-check/tests/q2.py 1
                        Failed example:
                            1 == 1
                        Expected:
                            False
                        Got:
                            True"""), 
                    "Did not pass correct tests"
                )


    def test_notebook_class(self):
        """
        Checks that the otter.Notebook class works correctly
        """
        grader = Notebook(TEST_FILES_PATH + "tests")

        def square(x):
            return x**2

        def negate(x):
            return not x

        global_env = {
            "square" : square,
            "negate" : negate
        }

        for q_path in glob(TEST_FILES_PATH + "tests/*.py"):
            q = os.path.split(q_path)[1][:-3]
            result = grader.check(q, global_env=global_env)
            if q != "q2":
                self.assertEqual(result.grade, 1, "Test {} failed".format(q))
            else:
                self.assertEqual(result.grade, 0, "Test {} passed".format(q))
    
    @classmethod
    def tearDownClass(cls):
        if os.path.exists(".OTTER_LOG"):
            os.system("rm .OTTER_LOG")



"""
In-container grading interface for Otter Grader
"""

import argparse
import os
import re
import json
import zipfile
import shutil
import tempfile
import nb2pdf
import pandas as pd

from glob import glob
try:
    from IPython.core.inputsplitter import IPythonInputSplitter
except ImportError:
    raise ImportError('IPython needs to be installed for notebook grading')

from ..execute import grade_notebook
from ..utils import id_generator

def grade(ipynb_path, pdf, script, ignore_errors=True, seed=None, cwd=None):
    """
    Grades a single ipython notebook and returns the score

    If no PDF is needed, set the pdf, tag_filter, and html_filter parameters to false. For .py
    files, set script to true.

    Args:
        ipynb_path (``str``): path to the notebook
        pdf (``bool``): whether unfiltered PDFs of notebooks should be generated
        tag_filter (``bool``): whether cell tag-filtered PDFs of notebooks should be generated
        html_filter (``bool``): whether HTML comment-filtered PDFs of notebooks should be generated
        script (``bool``): whether the input file is a Python script
        ignore_errors (``bool``, optional): whether errors should be ignored during execution
        seed (``int``, optional): random seed for intercell seeding
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment

    Returns:
        ``dict``: a score mapping with keys for each test, the student's scores, and total points 
            earned and possible 
    """
    # # get path of notebook file
    # base_path = os.path.dirname(ipynb_path)

    # glob tests
    test_files = glob('/home/tests/*.py')

    # get score
    result = grade_notebook(ipynb_path, test_files, script=script, ignore_errors=ignore_errors, seed=seed, cwd=cwd)

    # output PDF
    if pdf:
        nb2pdf.convert(
            ipynb_path, 
            filtering = pdf != "unfiltered", 
            filter_type = pdf if pdf != "unfiltered" else "tags",
            timeout=60000
        )

    return result

def main(args=None):
    """
    Parses command line arguments and executes submissions. Writes grades to a CSV file and optionally
    generates PDFs of submissions.

    Args:
        args (``list`` of ``str``, optional): alternate command line arguments
    """
    # implement argparser
    parser = argparse.ArgumentParser()
    parser.add_argument('submission_directory', help="Path to submissions directory")
    parser.add_argument("--pdfs", default=False, const="unfiltered", choices=["unfiltered", "tags", "html"], nargs="?")
    parser.add_argument("--scripts", action="store_true", default=False)
    parser.add_argument("--seed", type=int, default=None, help="A random seed to be executed before each cell")
    parser.add_argument("--zips", action="store_true", default=False, help="Whether the submissions are Notebook.export zip archives")
    parser.add_argument("--verbose", default=False, action="store_true", help="If present prints scores and hints to stdout")
    parser.add_argument("--debug", default=False, action="store_true", help="Does not ignore errors on execution")

    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    # cd into dir containing submissions
    subs_dir = os.path.abspath(args.submission_directory)
    os.chdir(subs_dir)

    # get file extension and get all of those files in dir_path
    file_extension = (".zip", ".py", ".ipynb")[[args.zips, args.scripts, True].index(True)]
    files = [(f, os.path.join(subs_dir, f)) for f in os.listdir(subs_dir) if os.path.splitext(f)[1] == file_extension]

    all_results = {"file": [], "score": [], "manual": []}

    if not args.pdfs:
        del all_results["manual"]

    # if zips, creat subdir to grade in
    if file_extension == ".zip":
        secret = id_generator()
        grading_dir = os.path.join(subs_dir, f"SUBMISSION_GRADING_DIRECTORY_{secret}")

        # copy all support files/folders into a template grading dir which we will copy from 
        # each time we grade a submission
        template_grading_dir = tempfile.mkdtemp()
        for fn in os.listdir(subs_dir):
            fp = os.path.join(subs_dir, fn)
            if os.path.isfile(fp) and os.path.splitext(fp)[1] != file_extension:
                shutil.copy(fp, template_grading_dir)
            elif os.path.isdir(fp):
                shutil.copytree(fp, os.path.join(template_grading_dir, fn))
    
    else:
        grading_dir = None

    # grade files
    for fname, fpath in files:
        os.chdir(subs_dir)

        all_results["file"].append(fname)

        # if zips, extract zip file into grading dir and find the gradeable filename
        if file_extension == ".zip":
            shutil.copytree(template_grading_dir, grading_dir)
            zf = zipfile.ZipFile(fpath)
            zf.extractall(grading_dir)
            os.chdir(grading_dir)
            gradeables = glob(os.path.join(grading_dir, ("*.py", "*.ipynb")[not args.scripts]))
            assert len(gradeables) == 1, f"Wrong number of gradeable files in submission {fname}"
            fpath = gradeables[0]

        # grade the submission
        score = grade(
            fpath, 
            args.pdfs, 
            args.scripts, 
            ignore_errors = not args.debug, 
            seed = args.seed, 
            cwd = grading_dir if grading_dir else subs_dir
        )
        score = score.to_dict()

        if args.verbose:
            print("Score details for {}".format(fname))
            print(json.dumps(score, default=lambda o: repr(o), indent=2))

        all_results["score"].append({t : score[t]["score"] if type(score[t]) == dict else score[t] for t in score})
        
        if args.pdfs:
            pdf_path = os.path.splitext(fpath)[0] + ".pdf"
            all_results["manual"].append(pdf_path)

        if file_extension == ".zip":
            shutil.rmtree(grading_dir)

    os.chdir(subs_dir)

    if file_extension ==  ".zip":
        shutil.rmtree(template_grading_dir)

    try:
        # expand mappings in all_results["score"]
        for q in all_results["score"][0].keys():
            all_results[q] = []

        for score in all_results["score"]:
            for q in score:
                all_results[q] += [score[q]]#["score"]]

    except IndexError:
        pass

    del all_results["score"]

    final_results = pd.DataFrame(all_results)
    final_results.to_csv("grades.csv", index=False)

if __name__ == "__main__":
    main()



"""
Utilities for Otter-Grader
"""

import os
import sys
import pathlib
import random
import string

from contextlib import contextmanager, redirect_stdout
from IPython import get_ipython

@contextmanager
def block_print():
    """
    Context manager that disables printing to stdout.
    """
    with open(os.devnull, "w") as f, redirect_stdout(f):
        yield

def flush_inline_matplotlib_plots():
    """
    Flush matplotlib plots immediately, rather than asynchronously
    
    Basically, the inline backend only shows the plot after the entire cell executes, which means we 
    can't easily use a context manager to suppress displaying it. See https://github.com/jupyter-widgets/ipywidgets/issues/1181/ 
    and https://github.com/ipython/ipython/issues/10376 for more details. This function displays flushes 
    any pending matplotlib plots if we are using the inline backend. Stolen from 
    https://github.com/jupyter-widgets/ipywidgets/blob/4cc15e66d5e9e69dac8fc20d1eb1d7db825d7aa2/ipywidgets/widgets/interaction.py#L35
    """
    if 'matplotlib' not in sys.modules:
        # matplotlib hasn't been imported, nothing to do.
        return

    try:
        import matplotlib as mpl
        from ipykernel.pylab.backend_inline import flush_figures
    except ImportError:
        return
    # except KeyError:
    #     return

    if mpl.get_backend() == 'module://ipykernel.pylab.backend_inline':
        flush_figures()

@contextmanager
def hide_outputs():
    """
    Context manager for hiding outputs from ``display()`` calls. IPython handles matplotlib outputs 
    specially, so those are supressed too.
    """
    ipy = get_ipython()
    if ipy is None:
        # Not running inside ipython!
        yield
        return
    old_formatters = ipy.display_formatter.formatters
    ipy.display_formatter.formatters = {}
    try:
        yield
    finally:
        # flush_inline_matplotlib_plots()
        ipy.display_formatter.formatters = old_formatters

def id_generator(size=6, chars=string.ascii_uppercase + string.digits):
    """Used to generate a dynamic variable name for grading functions

    This function generates a random name using the given length and character set.
    
    Args:
        size (``int``): length of output name
        chars (``str``, optional): set of characters used to create function name
    
    Returns:
        ``str``: randomized string name for grading function
    """
    return ''.join(random.choice(chars) for _ in range(size))

def get_variable_type(obj):
    """Returns the fully-qualified type string of an object ``obj``

    Args:
        obj (object): the object in question

    Returns:
        ``str``: the fully-qualified type string
    """
    return type(obj).__module__ + "." + type(obj).__name__

def get_relpath(src, dst):
    """
    Returns the relative path from ``src`` to ``dst``

    Args:
        src (``pathlib.Path``): the source directory
        dst (``pathlib.Path``): the destination directory

    Returns:
        ``pathlib.Path``: the relative path
    """
    # osrc = src
    ups = 0
    while True:
        try:
            dst.relative_to(src)
            break
        except ValueError:
            src = src.parent
            ups += 1
    return pathlib.Path(("../" * ups) / dst.relative_to(src))

def get_source(cell):
    """
    Returns the source code of a cell in a way that works for both nbformat and JSON
    
    Args:
        cell (``nbformat.NotebookNode``): notebook cell
    
    Returns:
        ``list`` of ``str``: each line of the cell source stripped of ending line breaks
    """
    source = cell['source']
    if isinstance(source, str):
        return cell['source'].split('\n')
    elif isinstance(source, list):
        return [line.strip('\n') for line in source]
    assert 'unknown source type', type(source)



"""
IPython notebook PDF Exporter forked from nb2pdf and gsExport
"""

import os
import warnings
import nbformat
import nbconvert
import pkg_resources

from .filter import load_notebook

def notebook_to_pdf(nb, dest, templating="test.tplx", save_tex=False, debug=False):
    """
    Writes a parsed notebook ``nb`` to a PDF file using nbconvert. Forked from
    https://github.com/dibyaghosh/gsExport/blob/master/gsExport/utils.py
    
    Args:
        nb (``nbformat.NotebookNode``): parsed notebook
        dest (``str``): path to write PDF
        templating (``str``, optional): path to template file relative to install directory of Otter
        save_tex (``bool``, optional): whether to save the LaTeX file as well
        debug (``bool``, optional): whether to run export in debug mode
    """
    warnings.filterwarnings("ignore", r"invalid escape sequence '\\c'", DeprecationWarning)

    if save_tex:
        latex_exporter = nbconvert.LatexExporter()

    pdf_exporter = nbconvert.PDFExporter()
    pdf_exporter.template_file = pkg_resources.resource_filename(__name__, templating)
    try:
        pdf_output = pdf_exporter.from_notebook_node(nb)
        with open(dest, "wb") as output_file:
            output_file.write(pdf_output[0])
        if save_tex:
            latex_output = latex_exporter.from_notebook_node(nb)
            with open(os.path.splitext(dest)[0] + ".tex", "w+") as output_file:
                output_file.write(latex_output[0])
    except nbconvert.pdf.LatexFailed as error:
        print("There was an error generating your LaTeX")
        output = error.output
        if debug:
            print("Showing full error message from PDFTex")

        else:
            print("Showing concise error message")
            output = "\n".join(error.output.split("\n")[-15:])
        print("=" * 60)
        print(output)
        print("=" * 60)

def export_notebook(nb_path, dest=None, filtering=False, pagebreaks=False, save_tex=False, debug=False):
    """
    Exports a notebook file at ``nb_path`` to a PDF with optional filtering and pagebreaks

    Args:
        nb_path (``str``): path to notebook
        dest (``str``, optional): path to write PDF
        filtering (``bool``, optional): whether the PDF should be filtered with HTML comments
        pagebreaks (``bool``, optional): whether there should be pagebreaks after questions in the PDF
        save_tex (``bool``, optional): whether to save the LaTeX file as well
        debug (``bool``, optional): whether to run export in debug mode
    """
    notebook = load_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)

    if dest is not None:
        pdf_name = dest
    else:
        pdf_name = os.path.splitext(nb_path)[0] + ".pdf"
        
    notebook_to_pdf(notebook, pdf_name, save_tex=save_tex, debug=debug)

def main(args):
    """
    Runs Otter Export

    Args:
        args (``argparse.Namespace``): parsed command line arguments
    """
    export_notebook(
        args.source,
        dest = args.dest,
        filtering = args.filtering,
        pagebreaks = args.pagebreaks,
        save_tex = args.save_tex,
        debug = args.debug
    )



"""
Gradescope autograder configuration generator for Otter Generate
"""

import os
import shutil
import subprocess
import pathlib
import pkg_resources

from glob import glob
from subprocess import PIPE
from jinja2 import Template

from .token import APIClient

TEMPLATES_DIR = pkg_resources.resource_filename(__name__, "templates")
SETUP_SH_PATH = os.path.join(TEMPLATES_DIR, "setup.sh")
REQUIREMENTS_PATH = os.path.join(TEMPLATES_DIR, "requirements.txt")
RUN_AUTOGRADER_PATH = os.path.join(TEMPLATES_DIR, "run_autograder")
MINICONDA_INSTALL_URL = "https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh"

with open(SETUP_SH_PATH) as f:
    SETUP_SH = Template(f.read())

with open(REQUIREMENTS_PATH) as f:
    REQUIREMENTS = Template(f.read())

with open(RUN_AUTOGRADER_PATH) as f:
    RUN_AUTOGRADER = Template(f.read())

def main(args):
    """
    Runs ``otter generate autograder``
    """
    assert args.threshold is None or 0 <= args.threshold <= 1, "{} is not a valid threshold".format(
        args.threshold
    )

    if args.course_id or args.assignment_id:
        assert args.course_id and args.assignment_id, "Either course ID or assignment ID unspecified for PDF submissions"
        if not args.token:
            args.token = APIClient.get_token()

    # check that args.public_multiplier is valid
    assert 0 <= args.public_multiplier <= 1, f"Public test multiplier {args.public_multiplier} is not between 0 and 1"

    # format run_autograder
    run_autograder = RUN_AUTOGRADER.render(
        threshold = str(args.threshold),
        points = str(args.points),
        show_stdout = str(args.show_stdout),
        show_hidden = str(args.show_hidden),
        seed = str(args.seed),
        token = str(args.token),
        course_id = str(args.course_id),
        assignment_id = str(args.assignment_id),
        filtering = str(not args.unfiltered_pdfs),
        pagebreaks = str(not args.no_pagebreaks),
        grade_from_log = str(args.grade_from_log),
        serialized_variables = str(args.serialized_variables),
        public_multiplier = str(args.public_multiplier),
        autograder_dir = str(args.autograder_dir),
    )

    setup_sh = SETUP_SH.render(
        miniconda_install_url = MINICONDA_INSTALL_URL
    )

    # create tmp directory to zip inside
    os.mkdir("./tmp")

    try:
        # copy tests into tmp
        os.mkdir(os.path.join("tmp", "tests"))
        for file in glob(os.path.join(args.tests_path, "*.py")):
            shutil.copy(file, os.path.join("tmp", "tests"))

        if os.path.isfile(args.requirements):
            with open(args.requirements) as f:
                requirements = REQUIREMENTS.render(
                    overwrite = args.overwrite_requirements,
                    other_requirements = f.read()
                )
        elif args.requirements != "requirements.txt":
            assert False, "requirements file {} not found".format(args.requirements)
        else:
            requirements = REQUIREMENTS.render(
                overwrite = args.overwrite_requirements,
                other_requirements = ""
            )

        # copy requirements into tmp
        with open(os.path.join(os.getcwd(), "tmp", "requirements.txt"), "w+") as f:
            f.write(requirements)

        # write setup.sh and run_autograder to tmp
        with open(os.path.join(os.getcwd(), "tmp", "setup.sh"), "w+") as f:
            f.write(setup_sh)

        with open(os.path.join(os.getcwd(), "tmp", "run_autograder"), "w+") as f:
            f.write(run_autograder)

        # copy files into tmp
        if len(args.files) > 0:
            os.mkdir(os.path.join("tmp", "files"))

            for file in args.files:
                # if a directory, copy the entire dir
                if os.path.isdir(file):
                    shutil.copytree(file, str(os.path.join("tmp", "files", os.path.basename(file))))
                else:
                    # check that file is in subdir
                    file = os.path.abspath(file)
                    assert os.getcwd() in file, \
                        f"{file} is not in a subdirectory of the working directory"
                    wd_path = pathlib.Path(os.getcwd())
                    file_path = pathlib.Path(file)
                    rel_path = file_path.parent.relative_to(wd_path)
                    output_path = os.path.join("tmp", "files", rel_path)
                    os.makedirs(output_path, exist_ok=True)
                    shutil.copy(file, output_path)

        os.chdir("./tmp")

        zip_path = os.path.join("..", args.output_path, "autograder.zip")
        if os.path.exists(zip_path):
            os.remove(zip_path)

        zip_cmd = ["zip", "-r", zip_path, "run_autograder",
                "setup.sh", "requirements.txt", "tests"]

        if args.files:
            zip_cmd += ["files"]

        zipped = subprocess.run(zip_cmd, stdout=PIPE, stderr=PIPE)

        if zipped.stderr.decode("utf-8"):
            raise Exception(zipped.stderr.decode("utf-8"))

        # move back to tmp parent directory
        os.chdir("..")
    
    except:
        # delete tmp directory
        shutil.rmtree("tmp")

        # raise the exception
        raise
    
    else:
        # delete tmp directory
        shutil.rmtree("tmp")



"""
Constants (global variables) for Otter Assign
"""

import re

NB_VERSION = 4
BLOCK_QUOTE = "```"
COMMENT_PREFIX = "#"
TEST_HEADERS = ["TEST", "HIDDEN TEST"]
ALLOWED_NAME = re.compile(r'[A-Za-z][A-Za-z0-9_]*')
NB_VERSION = 4

TEST_REGEX = r"(##\s*(hidden\s*)?test\s*##|#\s*(hidden\s*)?test)"
MD_SOLUTION_REGEX = r"(<strong>|\*{2})solution:?(<\/strong>|\*{2})"
SEED_REGEX = r"##\s*seed\s*##"



"""
Gradescope autograding internals
"""

import json
import os
import shutil
import subprocess
import re
import pickle
import warnings
import pandas as pd

from glob import glob
from textwrap import dedent

from .token import APIClient
from .utils import replace_notebook_instances

from ..check.logs import Log, QuestionNotInLogException
from ..check.notebook import _OTTER_LOG_FILENAME
from ..execute import grade_notebook
from ..export import export_notebook
from ..version import LOGO_WITH_VERSION


def main(config):
    """
    Runs autograder on Gradescope based on predefined configurations.

    Args:
        config (``dict``): configurations for autograder
    """
    print(LOGO_WITH_VERSION, "\n")
    
    os.chdir(config.get("autograder_dir", "/autograder"))

    if config.get("token", None) is not None:
        client = APIClient(token=config.get("token", None))
        generate_pdf = True
    else:
        generate_pdf = False

    # put files into submission directory
    if os.path.exists("./source/files"):
        for file in os.listdir("./source/files"):
            fp = os.path.join("./source/files", file)
            if os.path.isdir(fp):
                if not os.path.exists(os.path.join("./submission", os.path.basename(fp))):
                    shutil.copytree(fp, os.path.join("./submission", os.path.basename(fp)))
            else:
                shutil.copy(fp, "./submission")

    # create __init__.py files
    subprocess.run(["touch", "./__init__.py"])
    subprocess.run(["touch", "./submission/__init__.py"])

    os.makedirs("./submission/tests", exist_ok=True)
    tests_glob = glob("./source/tests/*.py")
    for file in tests_glob:
        shutil.copy(file, "./submission/tests")

    os.chdir("./submission")

    nb_path = glob("*.ipynb")[0]

    replace_notebook_instances(nb_path)

    # if glob("*.otter"):
    #     assert len(glob("*.otter")) == 1, "Too many .otter files (max 1 allowed)"
    #     with open(glob("*.otter")[0]) as f:
    #         otter_config = json.load(f)
    # else:
    #     otter_config = None

    if os.path.isfile(_OTTER_LOG_FILENAME):
        log = Log.from_file(_OTTER_LOG_FILENAME, ascending=False)
        if config.get("grade_from_log", False):
            print("\n\n")     # for logging in otter.execute.execute_log
    else:
        assert not config.get("grade_from_log", False), "missing log"
        log = None

    scores = grade_notebook(
        nb_path, 
        glob("./tests/*.py"), 
        name="submission", 
        cwd=".", 
        test_dir="./tests",
        ignore_errors=not config.get("debug", False), 
        seed=config.get("seed", None),
        log=log if config.get("grade_from_log", False) else None,
        variables=config.get("serialized_variables", None)
    )

    # verify the scores against the log
    print("\n\n")
    if log is not None:
        try:
            found_discrepancy = scores.verify_against_log(log)
            if not found_discrepancy:
                print("No discrepancies found while verifying scores against the log.")
        except BaseException as e:
            print(f"Error encountered while trying to verify scores with log:\n{e}")
    else:
        print("No log found with which to verify student scores")

    if generate_pdf:
        try:
            export_notebook(nb_path, filtering=config.get("filtering"), pagebreaks=config.get("pagebreaks"))
            pdf_path = os.path.splitext(nb_path)[0] + ".pdf"

            # get student email
            with open("../submission_metadata.json") as f:
                metadata = json.load(f)

            student_emails = []
            for user in metadata["users"]:
                student_emails.append(user["email"])
            
            for student_email in student_emails:
                client.upload_pdf_submission(config["course_id"], config["assignment_id"], student_email, pdf_path)

            print("\n\nSuccessfully uploaded submissions for: {}".format(", ".join(student_emails)))

        except:
            print("\n\n")
            warnings.warn("PDF generation or submission failed", RuntimeWarning)

    output = scores.to_gradescope_dict(config)

    os.chdir(config.get("autograder_dir", "/autograder"))

    with open("./results/results.json", "w+") as f:
        json.dump(output, f, indent=4)

    print("\n\n")
    df = pd.DataFrame(output["tests"])
    if "output" in df.columns:
        df.drop(columns=["output"], inplace=True)
    # df.drop(columns=["hidden"], inplace=True)
    print(df)


"""
Execution and grading internals for Otter-Grader
"""

import json
import itertools
import inspect

from IPython import get_ipython

from .execute_log import execute_log
from .execute_notebook import execute_notebook, filter_ignored_cells
from .execute_script import execute_script
from .results import GradingResults
from ..test_files import TestCollection, OKTestFile
from ..utils import id_generator

def check(test_file_path, global_env=None):
    """
    Checks ``global_env`` against given ``test_file`` in OK-format. If global_env is ``None``, the 
    global environment of the calling function is used. The following two calls are equivalent:

    .. code-block:: python
        check('tests/q1.py')
        check('tests/q1.py', globals())
    
    Args:
        test_file_path (``str``): path to test file
        global_env (``dict``, optional): a global environment resulting from the execution 
            of a python script or notebook

    Returns:
        ``otter.test_files.abstract_test.TestCollectionResults``: result of running the tests in the 
            given global environment

    """
    tests = TestCollection([test_file_path], OKTestFile)

    if global_env is None:
        # Get the global env of our callers - one level below us in the stack
        # The grade method should only be called directly from user / notebook
        # code. If some other method is calling it, it should also use the
        # inspect trick to pass in its parents' global env.
        global_env = inspect.currentframe().f_back.f_globals

    return tests.run(global_env, include_grade=False)

def grade_notebook(notebook_path, tests_glob=None, name=None, ignore_errors=True, script=False, 
    cwd=None, test_dir=None, seed=None, log=None, variables=None):
    """
    Grade a notebook file & return grade information

    This function grades a single Jupyter notebook using the provided tests. If grading a Python file,
    set ``script`` to true. 

    Args:
        notebook_path (``str``): path to a single notebook
        tests_glob (``list`` of ``str``, optional): names of test files
        name (``str``, optional): initial environment name
        ignore_errors (``bool``, optional): whether errors in execution should be ignored
        script (``bool``, optional): whether the notebook_path is a Python script
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        seed (``int``, optional): random seed for intercell seeding
        log (``otter.check.logs.Log``, optional): log from which to grade questions
        variables (``dict``, optional): map of variable names -> type string to check type of deserialized
            object to prevent arbitrary code from being put into the environment; ignored if log is ``None``

    Returns:
        ``otter.execute.results.GradingResults``: the results of grading
    """
    # ensure this is not being executed inside a notebook
    assert get_ipython() is None, "Cannot execute inside Jupyter Notebook"

    if not script:
        try:
            with open(notebook_path) as f:
                nb = json.load(f)
        except UnicodeDecodeError:
            with open(notebook_path, encoding='utf-8') as f:
                nb = json.load(f)
    else:
        with open(notebook_path) as f:
            nb = f.read()

    # remove any ignored cells from the notebook
    if not script:
        nb = filter_ignored_cells(nb)

    secret = id_generator()
    results_array = "check_results_{}".format(secret)
    initial_env = {
        results_array: []
    }

    if name:
        initial_env["__name__"] = name

    if log is not None:
        global_env = execute_log(nb, log, secret, initial_env, ignore_errors=ignore_errors, cwd=cwd, test_dir=test_dir, variables=variables)
    elif script:
        global_env = execute_script(nb, secret, initial_env, ignore_errors=ignore_errors, cwd=cwd, seed=seed)
    else:
        global_env = execute_notebook(nb, secret, initial_env, ignore_errors=ignore_errors, cwd=cwd, test_dir=test_dir, seed=seed)

    test_results = global_env[results_array]

    # Check for tests which were not included in the notebook and specified by tests_globs
    # Allows instructors to run notebooks with additional tests not accessible to user
    if tests_glob:
        # unpack list of paths into a single list
        tested_set = list(itertools.chain(*[r.paths for r in test_results]))
        extra_tests = []
        for t in sorted(tests_glob):
            include = True
            for tested in tested_set:
                if tested in t:     # e.g. if 'tests/q1.py' is in /srv/repo/lab01/tests/q1.py'
                    include = False
            if include:
                extra_tests.append(TestCollection([t], OKTestFile))
        extra_results = [t.run(global_env, include_grade=False) for t in extra_tests]
        test_results += extra_results

    return GradingResults(test_results)



"""
Otter Assign command-line utility
"""

import os
import pathlib
import nb2pdf

from .assignment import Assignment
from .output import write_output_directories
from .utils import run_tests, write_otter_config_file, run_generate_autograder

from .. import _WINDOWS
from ..export import export_notebook
from ..utils import get_relpath, block_print

# for now can't use nb2pdf on Windows b/c of pyppeteer - this may be due to my
# local install, requires further debugging
TO_PDF_FN = (nb2pdf.convert, export_notebook)[_WINDOWS]

def main(args):
    """
    Runs Otter Assign
    
    Args:
        ``argparse.Namespace``: parsed command line arguments
    """
    master, result = pathlib.Path(args.master), pathlib.Path(args.result)
    print("Generating views...")

    assignment = Assignment()
    
    # TODO: update this condition
    if True:
        result = get_relpath(master.parent, result)
        orig_dir = os.getcwd()
        os.chdir(master.parent)
        master = pathlib.Path(master.name)
    
    assignment.master, assignment.result = master, result
    
    try:
        write_output_directories(master, result, assignment, args)

        # check that we have a seed if needed
        if assignment.seed_required:
            assert not assignment.generate or assignment.generate.get('seed', None) is not None, \
                "Seeding cell found but no seed provided"
        
        # generate PDF of solutions with nb2pdf
        if assignment.solutions_pdf:
            print("Generating solutions PDF...")
            filtering = assignment.solutions_pdf == 'filtered'
            TO_PDF_FN(
                str(result / 'autograder' / master.name),
                dest=str(result / 'autograder' / (master.stem + '-sol.pdf')),
                filtering=filtering
            )

        # generate a tempalte PDF for Gradescope
        if assignment.template_pdf:
            print("Generating template PDF...")
            export_notebook(
                str(result / 'autograder' / master.name),
                dest=str(result / 'autograder' / (master.stem + '-template.pdf')), 
                filtering=True, 
                pagebreaks=True
            )

        # generate the .otter file if needed
        if assignment.service or assignment.save_environment:
            write_otter_config_file(master, result, assignment)

        # generate Gradescope autograder zipfile
        if assignment.generate:
            # TODO: move this to another function
            print("Generating autograder zipfile...")
            run_generate_autograder(result, assignment, args)

        # run tests on autograder notebook
        if assignment.run_tests and not args.no_run_tests:
            print("Running tests...")
            with block_print():
                if isinstance(assignment.generate, bool):
                    seed = None
                else:
                    seed = assignment.generate.get('seed', None)
                run_tests(result / 'autograder' / master.name, debug=args.debug, seed=seed)
            print("All tests passed!")
    
    # for tests
    except:
        # TODO: change this condition
        if True:
            os.chdir(orig_dir)

        raise

    else:
        # TODO: change this condition
        if True:
            os.chdir(orig_dir)



"""
Otter Service tornado server
"""

import os
import json
import yaml
import hashlib
import jwt
import tornado.options
import queries
import stdio_proxy

from io import BytesIO
from datetime import datetime
from binascii import hexlify
from tornado.httpserver import HTTPServer
from tornado.web import Application, RequestHandler
from tornado.auth import GoogleOAuth2Mixin
from tornado.ioloop import IOLoop
from tornado.queues import Queue
from concurrent.futures import ThreadPoolExecutor

from .utils import connect_db
from ..grade.containers import grade_assignments

OTTER_SERVICE_DIR = "/otter-service"
ARGS = None
SUBMISSION_QUEUE = Queue()
CONN = None
EXECUTOR = ThreadPoolExecutor()

class BaseHandler(tornado.web.RequestHandler):
    """Base login handler"""
    def get_current_user(self):
        """
        Gets secure user cookie for personal authentication
        """
        return self.get_secure_cookie("user")

class LoginHandler(BaseHandler):
    """
    Default auth handler
    
    A login handler that requires instructors to setup users and passwords in database beforehand, 
    allowing students to auth within the notebook.
    """
    async def get(self):
        """
        GET request handler for personal/default authentication login
        """
        username = self.get_argument('username', True)
        password = self.get_argument('password', True)
        pw_hash = hashlib.sha256(password.encode()).hexdigest()
        account_check = await self.db.query(
            """
            SELECT * FROM users 
            WHERE username = %s AND password = %s
            """,
            [username, pw_hash]
        )
        if len(account_check) > 0:
            print("Logged in user {} and generating API key".format(username))
            account_check.free()
            api_key = hexlify(os.urandom(32)).decode("utf-8")
            self.write(api_key)
            results = await self.db.query(
                """
                INSERT INTO users (api_keys, username, password) VALUES (%s, %s, %s)
                ON CONFLICT (username)
                DO UPDATE SET api_keys = array_append(users.api_keys, %s)
                """,
                [[api_key], username, pw_hash, api_key]
            )
            results.free()
        else:
            print("Failed login attempt for user {}".format(username))
            account_check.free()
            self.clear()
            self.set_status(401)
            self.finish()

    @property
    def db(self):
        return self.application.db

class GoogleOAuth2LoginHandler(RequestHandler, GoogleOAuth2Mixin):
    async def get(self):
        """
        GET request handler for Google OAuth

        Handler for authenticating users with Google OAuth. Requires that user sets environment
        variables containing their client key and secret. Provides users with an API key that they
        can enter in the notebook by way of authenticating.
        """
        if not self.get_argument('code', False):
            print("Redirecting user to Google OAuth")
            return self.authorize_redirect(
                redirect_uri=self.settings['auth_redirect_uri'],
                client_id = ARGS.google_key if ARGS.google_key else self.settings['google_oauth']['key'],
                client_secret = ARGS.google_secret if ARGS.google_secret else self.settings['google_oauth']['secret'],
                scope=['email', 'profile'],
                response_type='code',
                extra_params={'approval_prompt': 'auto'}
            )
        else:
            resp = await self.get_authenticated_user(
                redirect_uri=self.settings['auth_redirect_uri'],
                code=self.get_argument('code')
            )
            api_key = hexlify(os.urandom(32)).decode("utf-8")
            email = jwt.decode(resp['id_token'], verify=False)['email']
            print("Generating API key for user {} from Google OAuth".format(email))
            results = await self.db.query(
                """
                INSERT INTO users (api_keys, email) VALUES (%s, %s)
                ON CONFLICT (email) 
                DO UPDATE SET api_keys = array_append(users.api_keys, %s)
                """,
                [[api_key], email, api_key]
            )
            results.free()

            self.render("templates/api_key.html", key=api_key)

    @property
    def db(self):
        return self.application.db

class SubmissionHandler(RequestHandler):
    """
    Processes and validates student submission

    Handler for processing and validating a student's submission. Ensure that notebook is present
    and valid, checks API key, and implements rate limiting to prevent spamming the autograder.
    Queues submission for grading by ``EXECUTOR``.
    """
    async def get(self):
        """
        GET request handler. Route warns users that this is a POST-only route.
        """
        self.write("This is a POST-only route; you probably shouldn't be here.")
        self.finish()

    async def post(self):
        """
        POST request handler. Validates JSON params and queues submission for grading.
        """
        self.submission_id = None
        try:
            # check request params
            request = tornado.escape.json_decode(self.request.body)
            assert 'nb' in request.keys(), 'submission contains no notebook'
            assert 'api_key' in request.keys(), 'missing api key'

            notebook = request['nb']
            api_key = request['api_key']
            
            # run through submission
            await self.submit(notebook, api_key)
        except Exception as e:
            print(e)
        self.finish()

        # if submission successful, queue notebook for grading
        if self.submission_id is not None:
            SUBMISSION_QUEUE.put(self.submission_id)

    async def validate(self, notebook, api_key):
        """
        Ensures a submision is valid by checking user credentials, submission frequency, and
        validity of notebook file.

        Arguments:
            notebook (``dict``): notebook in JSON form
            api_key (``str``): API key generated during submission

        Returns:
            ``tuple``: submission information
        """
        # authenticate user with api_key
        results = await self.db.query("SELECT user_id, username, email FROM users WHERE %s=ANY(api_keys) LIMIT 1", [api_key])
        user_id = results.as_dict()['user_id'] if len(results) > 0 else None
        username = results.as_dict()['username'] or results.as_dict()['email'] if len(results) > 0 else None
        results.free()
        assert user_id, 'invalid API key: {}'.format(api_key)

        # get assignment and class id from notebook metadata
        assert all(key in notebook for key in ['metadata', 'nbformat', 'cells']), 'invalid Jupyter notebook'
        assert 'assignment_id' in notebook['metadata'], 'missing required metadata attribute: assignment_id'
        assert 'class_id' in notebook['metadata'], 'missing required metadata attribute: class_id'
        assignment_id = notebook['metadata']['assignment_id']
        class_id = notebook['metadata']['class_id']

        # rate limiting
        results = await self.db.query(
            """
            SELECT timestamp 
            FROM submissions 
            WHERE user_id = %s AND assignment_id = %s AND class_id = %s
            ORDER BY timestamp DESC 
            LIMIT 1
            """, 
            (user_id, assignment_id, class_id)
        )
        last_submitted = results.as_dict()['timestamp'] if len(results) > 0 else None
        results.free()

        if last_submitted:
            delta = datetime.utcnow() - last_submitted
            # rate_limit = 120
            if delta.seconds < ARGS.rate_limit:
                self.write_error(429, message='Please wait {} second(s) before re-submitting.'.format(ARGS.rate_limit - delta.seconds))
                return
        
        # check that assignment exists
        results = await self.db.query("SELECT * FROM assignments WHERE assignment_id=%s LIMIT 1", [assignment_id])
        assert results, 'assignment_id {} not found on server'.format(assignment_id)
        assignment = results.as_dict()
        results.free()

        return (user_id, username, assignment['class_id'], assignment_id, assignment['assignment_name'])

    async def submit(self, notebook, api_key):
        """
        If valid submission, inserts notebook into submissions table in database and queues it 
        for grading.

        Arguments:
            notebook (``dict``): notebook in JSON form
            api_key (``str``): API key generated during submission
        """
        try:
            user_id, username, class_id, assignment_id, assignment_name = await self.validate(notebook, api_key)
        except TypeError as e:
            print("Submission failed for user with API key {}: ".format(api_key, e))
            return
        except AssertionError as e:
            print("Submission failed for user with API key {} due to due to client error: {}".format(api_key, e))
            self.write_error(400, message=e)
            return

        # fetch next submission id
        results = await self.db.query("SELECT nextval(pg_get_serial_sequence('submissions', 'submission_id')) as id")
        submission_id = results.as_dict()['id']
        results.free()

        print("Successfully received submission {} from user {}".format(submission_id, username))

        # save notebook to disk
        dir_path = os.path.join(
            self.settings['notebook_dir'],
            'class-{}'.format(class_id),
            'assignment-{}'.format(assignment_id),
            'submission-{}'.format(submission_id)
        )
        file_path = os.path.join(dir_path, '{}.ipynb'.format(assignment_name))
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        with open(file_path, 'w') as f:
            json.dump(notebook, f)

        print("Successfully saved submission {} at {}".format(submission_id, file_path))
        
        # store submission to database
        results = await self.db.query("INSERT INTO submissions (submission_id, assignment_id, class_id, user_id, file_path, timestamp) VALUES (%s, %s, %s, %s, %s, %s)",
                                            [submission_id, assignment_id, class_id, user_id, file_path, datetime.utcnow()])
        assert results, 'submission failed'
        results.free()

        self.submission_id = submission_id

        self.write('Submission {} received.'.format(submission_id))

    @property
    def db(self):
        return self.application.db

    def write_error(self, status_code, **kwargs):
        """
        Writes an error message to response

        Args:
            status_code (``int``): the response status
            message (``str``): message to include in the response
        """
        if 'message' in kwargs:
            self.write('Submission failed: {}'.format(kwargs['message']))
        else:
            self.write('Submission failed.')

def grade_submission(submission_id):
    """
    Grades a single submission with id ``submission_id``

    Args:
        submission_id (``str``): the id of the submission to grade

    Returns:
        ``tuple``: grading message and results dataframe for printing
    """
    global CONN
    cursor = CONN.cursor()

    cursor.execute(
        """
        SELECT user_id, submission_id, assignment_id, class_id, file_path 
        FROM submissions 
        WHERE submission_id = %s 
        ORDER BY timestamp DESC
        LIMIT 1
        """,
        (submission_id, )
    )
    user_record = cursor.fetchall()
    assert len(user_record) == 1, "Submission {} not found".format(submission_id)
    row = user_record[0]
    user_id = int(row[0])
    submission_id = int(row[1])
    assignment_id = str(row[2])
    class_id = str(row[3])
    file_path = str(row[4])

    cursor.execute(
        """
        SELECT seed
        FROM assignments
        WHERE assignment_id = %s AND class_id = %s
        """,
        (assignment_id, class_id)
    )
    assignment_record = cursor.fetchall()
    assert len(assignment_record) == 1, "Assignment {} for class {} not found".format(assignment_id, class_id)
    seed = int(assignment_record[0][0]) if assignment_record[0][0] else None

    cursor.execute(
        """
        SELECT username, email 
        FROM users 
        WHERE user_id = %s
        LIMIT 1
        """,
        (user_id, )
    )
    user_record = cursor.fetchall()
    row = user_record[0]
    username = str(row[0] or row[1])

    # Run grading function in a docker container
    stdout = BytesIO()
    stderr = BytesIO()
    try:
        with stdio_proxy.redirect_stdout(stdout), stdio_proxy.redirect_stderr(stderr):
            df = grade_assignments(
                tests_dir=None, 
                notebooks_dir=file_path, 
                id=assignment_id, 
                image=class_id + "-" + assignment_id,
                debug=True,
                verbose=True,
                seed=seed
            )
            
        message = "Graded submission {} from user {}".format(submission_id, username)

        df_json_str = df.to_json()
    
        # Insert score into submissions table
        cursor.execute(
            """
            UPDATE submissions
            SET score = %s
            WHERE submission_id = %s
            """,
            (df_json_str, submission_id)
        )

    finally:
        stdout = stdout.getvalue().decode("utf-8")
        stderr = stderr.getvalue().decode("utf-8")
        with open(os.path.join(os.path.split(file_path)[0], "GRADING_STDOUT"), "w+") as f:
            f.write(stdout)
        with open(os.path.join(os.path.split(file_path)[0], "GRADING_STDERR"), "w+") as f:
            f.write(stderr)
    
    cursor.close()
    return message, df

async def start_grading_queue(shutdown=False):
    """
    Pops submission ids off ``SUBMISSION_QUEUE`` and sending them into ``EXECUTOR`` to be graded
    
    Args:
        shutdown (``bool``): whether or not to shutdown EXECUTOR after processing queue; default 
            ``False``
    """
    global SUBMISSION_QUEUE

    async for submission_id in SUBMISSION_QUEUE:
        future = EXECUTOR.submit(
            grade_submission,
            submission_id
        )
        future.add_done_callback(lambda f: print(f.result()[0], "\n", f.result()[1]))

        # Set task done in queue
        SUBMISSION_QUEUE.task_done()
    
    if shutdown:
        EXECUTOR.shutdown(wait=True)

class Application(tornado.web.Application):
    """
    Otter Service tornado application
    """
    def __init__(self):
        """
        Initialize tornado server for receiving/grading submissions
        """
        endpoint = ARGS.endpoint or os.environ.get("OTTER_ENDPOINT", None)
        assert endpoint is not None, "no endpoint address provided"
        assert os.path.isdir(OTTER_SERVICE_DIR), "{} does not exist".format(OTTER_SERVICE_DIR)
        settings = dict(
            google_oauth={
                "key": ARGS.google_key or os.environ.get("GOOGLE_CLIENT_KEY", None), 
                "secret": ARGS.google_secret or os.environ.get("GOOGLE_CLIENT_SECRET", None)
            },
            notebook_dir = os.path.join(OTTER_SERVICE_DIR, "submissions"),
            auth_redirect_uri = os.path.join(endpoint, "auth/callback")
        )
        handlers = [
            (r"/submit", SubmissionHandler),
            (r"/auth/google", GoogleOAuth2LoginHandler),
            (r"/auth/callback", GoogleOAuth2LoginHandler),
            (r"/auth", LoginHandler)
        ]
        tornado.web.Application.__init__(self, handlers, **settings)
        
        # Initialize database session
        self.db = queries.TornadoSession(queries.uri(
            host=ARGS.db_host,
            port=ARGS.db_port,
            dbname='otter_db',
            user=ARGS.db_user,
            password=ARGS.db_pass
        ))

def main(cli_args):
    """
    Starts Otter Service tornado server

    Args:
        cli_args (``argparse.Namespace``): parsed command-line arguments
    """
    # if args.missing_packages:
    #     raise ImportError(
    #         "Missing some packages required for otter service. "
    #         "Please install all requirements at "
    #         "https://raw.githubusercontent.com/ucbds-infra/otter-grader/master/requirements.txt"
    #     )

    global CONN
    global ARGS

    ARGS = cli_args
    CONN = connect_db(ARGS.db_host, ARGS.db_user, ARGS.db_pass, ARGS.db_port)

    port = ARGS.port
    tornado.options.parse_command_line()

    # make submissions forlder
    if not os.path.isdir(OTTER_SERVICE_DIR):
        os.makedirs(os.path.join(OTTER_SERVICE_DIR))
    
    server = HTTPServer(Application())
    server.listen(port)
    print("Listening on port {}".format(port))

    IOLoop.current().add_callback(start_grading_queue)
    IOLoop.current().start()



####################################
##### Tests for otter generate #####
####################################

import os
import unittest
import subprocess
import json

from subprocess import PIPE
from glob import glob
from unittest import mock

from otter.argparser import get_parser
from otter.generate.autograder import main as autograder

from .. import TestCase

parser = get_parser()

TEST_FILES_PATH = "test/test_generate/test-autograder/"

class TestAutograder(TestCase):
    
    def create_docker_image(self):
        create_image_cmd = ["make", "docker-test"]
        create_image = subprocess.run(create_image_cmd, stdout=PIPE, stderr=PIPE)
        assert not create_image.stderr, create_image.stderr.decode("utf-8")

        # use docker image inspect to see that the image is installed and tagged as otter-grader
        inspect = subprocess.run(["docker", "image", "inspect", "otter-test"], stdout=PIPE, stderr=PIPE)

        # assert that it didn't fail, it will fail if it is not installed
        self.assertEqual(len(inspect.stderr), 0, inspect.stderr.decode("utf-8"))
   
    def test_gs_generator(self):
        """
        Check that the correct zipfile is created by gs_generator.py
        """
        # create the zipfile
        generate_command = ["generate", "autograder",
            "-t", TEST_FILES_PATH + "tests",
            "-o", TEST_FILES_PATH,
            "-r", TEST_FILES_PATH + "requirements.txt",
            TEST_FILES_PATH + "data/test-df.csv"
        ]
        args = parser.parse_args(generate_command)
        args.func = autograder
        args.func(args)

        # unzip the zipfile
        unzip_command = ["unzip", "-o", TEST_FILES_PATH + "autograder.zip", "-d", TEST_FILES_PATH + "autograder"]
        unzip = subprocess.run(unzip_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(unzip.stderr), 0, unzip.stderr)

        # # go through files and ensure that they are correct
        # for file in glob(TEST_FILES_PATH + "autograder/*"):
        #     if os.path.isfile(file):
        #         correct_file_path = os.path.join(TEST_FILES_PATH + "autograder-correct", os.path.split(file)[1])
        #         with open(file) as f:
        #             with open(correct_file_path) as g:
        #                 self.assertEqual(f.read(), g.read(), "{} does not match {}".format(file, correct_file_path))
        #     else:
        #         for subfile in glob(os.path.join(file, "*")):
        #             correct_file_path = os.path.join(TEST_FILES_PATH + "autograder-correct", os.path.split(file)[1], os.path.split(subfile)[1])
        #             with open(subfile) as f:
        #                 with open(correct_file_path) as g:
        #                     self.assertEqual(f.read(), g.read(), "{} does not match {}".format(subfile, correct_file_path))
        self.assertDirsEqual(TEST_FILES_PATH + "autograder", TEST_FILES_PATH + "autograder-correct")

        # cleanup files
        cleanup_command = ["rm", "-rf", TEST_FILES_PATH + "autograder", TEST_FILES_PATH + "autograder.zip"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(cleanup.stderr), 0, cleanup.stderr.decode("utf-8"))

    def test_gradescope(self):
        """
        Checks that the Gradescope autograder works
        """
        self.create_docker_image()

        # generate the zipfile
        generate_command = ["generate", "autograder",
            "-t", TEST_FILES_PATH + "tests",
            "-o", TEST_FILES_PATH,
            "-r", TEST_FILES_PATH + "requirements.txt",
            TEST_FILES_PATH + "data/test-df.csv"
        ]
        args = parser.parse_args(generate_command)
        args.func = autograder
        args.func(args)

        # build the docker image
        build = subprocess.run(["docker", "build", TEST_FILES_PATH, "-t", "otter-gradescope-test"], stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(build.stderr), 0, build.stderr.decode("utf-8"))

        # launch the container and return its container ID
        launch = subprocess.run(["docker", "run", "-dt", "otter-gradescope-test", "/autograder/run_autograder"], stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(launch.stderr), 0, launch.stderr.decode("utf-8"))

        # get container ID
        container_id = launch.stdout.decode("utf-8")[:-1]

        # attach to the container and wait for it to finish
        attach = subprocess.run(["docker", "attach", container_id], stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(attach.stderr), 0, attach.stderr.decode("utf-8"))
        
        # copy out the results.json file
        copy = subprocess.run(["docker", "cp", "{}:/autograder/results/results.json".format(container_id), "test/results.json"], stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(copy.stderr), 0, copy.stderr.decode("utf-8"))

        # check that we got the right results
        with open(TEST_FILES_PATH + "results-correct.json") as f:
            correct_results = json.load(f)
        with open("test/results.json") as f:
            results = json.load(f)
        self.assertEqual(results, correct_results, "Incorrect results when grading in Gradescope container")

        # cleanup files and container
        cleanup = subprocess.run(["rm", "-rf", TEST_FILES_PATH + "autograder", TEST_FILES_PATH + "autograder.zip", "test/results.json"], stdout=PIPE, stderr=PIPE)
        remove_container = subprocess.run(["docker", "rm", container_id], stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(cleanup.stderr), 0, cleanup.stderr.decode("utf-8"))
        self.assertEqual(len(remove_container.stderr), 0, remove_container.stderr.decode("utf-8"))



"""
Abstract base classes for working with test files and classes to represent collections of test and 
their results
"""

from abc import ABC, abstractmethod
from textwrap import dedent
from typing import Tuple, List, Dict, Any
from jinja2 import Template
from pygments import highlight
from pygments.lexers import PythonConsoleLexer
from pygments.formatters import HtmlFormatter

class TestFile(ABC):
    html_result_pass_template = Template("""
    <p><strong>{{ name }}</strong> passed!</p>
    """)

    plain_result_pass_template = Template("{{ name }} passed!")

    html_result_fail_template = Template("""
    <p><strong style='color: red;'>{{ name }}</strong></p>
    <p><strong>Test code:</strong><pre>{{test_code}}</pre></p>
    <p><strong>Test result:</strong><pre>{{test_result}}</pre></p>
    """)

    plain_result_fail_template = Template(dedent("""\
       {{ name }}

    Test result:
    {{test_result}}"""))

    def _repr_html_(self):
        if self.passed:
            return type(self).html_result_pass_template.render(name=self.name)
        else:
            return type(self).html_result_fail_template.render(
                name=self.name,
                test_code=highlight(self.failed_test, PythonConsoleLexer(), HtmlFormatter(noclasses=True)),
                test_result=self.result
            )

    def __repr__(self):
        if self.passed:
            return type(self).plain_result_pass_template.render(name=self.name)
        else:
            return type(self).plain_result_fail_template.render(
                name=self.name,
                test_code=self.failed_test, # highlight(self.failed_test, PythonConsoleLexer(), HtmlFormatter(noclasses=True)),
                test_result=self.result
            )

    # @abstractmethod
    def __init__(self, name, tests, hiddens, value=1, hidden=True):
        self.name = name
        self.public_tests = [t for t, h in zip(tests, hiddens) if not h]
        self.hidden_tests = [t for t, h in zip(tests, hiddens) if h]
        self.value = value
        self.hidden = hidden
        self.passed = None
        self.failed_test = None
        self.failed_test_hidden = None
        self.result = None

    @classmethod
    @abstractmethod
    def from_file(cls, path):
        ...

    @abstractmethod
    def run(self, global_environment):
        ...
    
class TestCollection:
    def __init__(self, test_paths: List[str], test_class: TestFile):
        self.paths = test_paths
        self.tests = [test_class.from_file(path) for path in self.paths if "__init__.py" not in path]
    
    def run(self, global_environment, include_grade=True):
        """Run this object's tests on a given global environment (from running a notebook/script)
        
        Arguments:
            global_environment (``dict``): result of executing a Python notebook/script
            include_grade (``bool``, optional): whether grade should be included in result
        
        Returns:
            ``TestCollectionResults``: object resulting from running tests on ``global_environment`` 
                with grade, tests passed, and more information
        """
        passed_tests = []
        failed_tests = []
        grade = 0
        total = 0
        for t in self.tests:
            total += t.value
            passed, test_obj = t.run(global_environment)
            if passed:
                grade += t.value
                passed_tests.append(t)
            else:
                failed_tests.append((t, test_obj))

        try:
            grade /= total
        except ZeroDivisionError:
            grade = 0

        return TestCollectionResults(
            grade, self.paths, self.tests, passed_tests, failed_tests, include_grade
        )

class TestCollectionResults:
    html_result_template = Template("""
    {% if include_grade %}
    <strong>Grade: {{ grade }}</strong>
    {% endif %}
    {% if grade == 1.0 %}
        <p>All tests passed!</p>
    {% else %}
        <p>{{ passed_tests|length }} of {{ tests|length }} tests passed</p>
        {% if passed_tests %}
        <p> <strong>Tests passed:</strong>
            {% for passed_test in passed_tests %} {{ passed_test.name }} {% endfor %}
        </p>
        {% endif %}
        {% if failed_tests %}
        <p> <strong>Tests failed: </strong>
            <ul>
            {% for failed_test, failed_test_obj in failed_tests %}
                <li> {{ failed_test_obj._repr_html_() }} </li>
            {% endfor %}
            </ul>
        {% endif %}
    {% endif %}
    """)

    plain_result_template = Template("""{% if include_grade %}Grade: {{ grade }}{% endif %}
    {% if grade == 1.0 %}All tests passed!{% else %}
    {{ passed_tests|length }} of {{ tests|length }} tests passed
    {% if passed_tests %}
    Tests passed: {% for passed_test in passed_tests %} {{ passed_test.name }} {% endfor %}{% endif %}
    {% if failed_tests %}
    Tests failed:
    {% for failed_test, failed_test_obj in failed_tests %}
        {{ failed_test_obj.__repr__() }}
    {% endfor %}
    {% endif %}
    {% endif %}
    """)

    def __init__(self, grade, paths, tests, passed_tests, failed_tests, include_grade=True):
        self.grade = grade
        self.paths = paths
        self.tests = tests
        self.passed_tests = passed_tests
        self.failed_tests = failed_tests
        self.include_grade = include_grade

    def _repr_html_(self):
        return type(self).html_result_template.render(
            grade=self.grade,
            passed_tests=self.passed_tests,
            failed_tests=self.failed_tests,
            tests=self.tests,
            include_grade=self.include_grade
        )

    def _repr_gradescope_(self):
        return type(self).plain_result_template.render(
            grade=self.grade,
            passed_tests=self.passed_tests,
            failed_tests=self.failed_tests,
            tests=self.tests,
            include_grade=self.include_grade
        )

    def __repr__(self):
        return type(self).plain_result_template.render(
            grade=self.grade,
            passed_tests=self.passed_tests,
            failed_tests=self.failed_tests,
            tests=self.tests,
            include_grade=self.include_grade
        )

    
    



"""
Cell filtering for IPython notebook exports forked from nb2pdf and gsExport
"""

import re
import nbformat

NBFORMAT_VERSION = 4
BEGIN_QUESTION_REGEX = r"<!--\s*BEGIN QUESTION\s*-->"
END_QUESTION_REGEX = r"<!--\s*END QUESTION\s*-->"
NEW_PAGE_REGEX = r"<!--\s*NEW PAGE\s*-->"
NEW_PAGE_CELL_SOURCE = "<!-- #newpage -->"

def load_notebook(nb_path, filtering=True, pagebreaks=False):
    """
    Loads notebook at ``nb_path`` with nbformat and returns the parsed notebook, optionally filtered
    and with pagebreak metadata hidden in HTML comments.

    Args:
        nb_path (``str``): path to notebook
        filtering (``bool``, optional): whetheer cells should be filtered
        pagebreaks (``bool``, optional): whether to include pagebreaks between each question; ignored
            if ``filtering`` is ``False``

    Returns:
        ``nbformat.NotebookNode``: the parsed and (optionally) filtered notebook
    """
    with open(nb_path) as f:
        notebook = nbformat.read(f, as_version=NBFORMAT_VERSION)
    if filtering:
        notebook = filter_notebook_cells_by_comments(notebook, pagebreaks=pagebreaks)
    return notebook

def has_begin(line):
    """
    Returns whether a string contains a begin question comment

    A begin question comment is an HTML comment on a single line that denotes the beginning of an 
    export block. The begin question comment looks like:

    .. code-block:: html

        <!-- BEGIN QUESTION -->
    
    Args:
        line (``str``): the line to search
    
    Returns:
        ``bool``: whether the line contains a substring matching the begin question regex
    """
    return bool(re.search(BEGIN_QUESTION_REGEX, line, flags=re.IGNORECASE))

def has_end(line):
    """
    Returns whether a string contains an end question comment

    An end question comment is an HTML comment on a single line that denotes the end of an export 
    block. The begin question comment looks like:

    .. code-block:: html

        <!-- END QUESTION -->
    
    Args:
        line (``str``): the line to search
    
    Returns:
        ``bool``: whether the line contains a substring matching the end question regex
    """
    return bool(re.search(END_QUESTION_REGEX, line, flags=re.IGNORECASE))

def sub_end_for_new_page(line):
    """
    Subsitutes an end question comment for a newpage comment

    The end question HTML comment (cf. ``otter.export.filter.has_end``) is replaced with the following
    HTML comment to indicate a pagebreak in the LaTeX template.

    .. code-block:: html

        <!-- #newpage -->

    Args:
        ``line``: the line to substitute in

    Returns:
        ``str``: the line with the end question match substituted for the newpage comment
    """
    return re.sub(END_QUESTION_REGEX, NEW_PAGE_CELL_SOURCE, line)

def filter_notebook_cells_by_comments(notebook, pagebreaks=False):
    """
    Filters a parsed notebook using HTML comments in Markdown cells. Optionally inserts pagebreak
    metadata as HTML comments.

    Args:
        notebook (``nbformat.NotebookNode``): the parsed notebook
        pagebreaks (``bool``, optional): whether to include pagebreaks between questions

    Returns:
        ``nbformat.NotebookNode``: the filtered notebook with (optional) pagebreaks
    """
    cells = notebook["cells"]

    idx_to_delete, in_question = [], False
    for curr_idx, cell in enumerate(cells):

        if isinstance(cell["source"], str):
            source = cell["source"].split("\n")
        elif isinstance(cell["source"], list):
            source = [line.strip("\n") for line in cell["source"]]
        else:
            raise Exception("Invalid notebook cell source: {}".format(cell))

        lines_before_begin, lines_after_end = -1, -1
        for line_idx, line in enumerate(source):

            # check for begin question regex in source
            if not in_question:
                if has_begin(line) and lines_before_begin == -1:
                    lines_before_begin = line_idx
                    in_question = True
            
            # check for end question regex in source
            else:
                if has_end(line):
                    lines_after_end = line_idx
                    in_question = False
                    
                    # if we are creating pagebreaks, sub the end regex for a newpage directive
                    if pagebreaks:
                        source[line_idx] = sub_end_for_new_page(line)

        # if we are not in question and there is no begin/end comment, delete the cell
        if lines_before_begin == -1 and lines_after_end == -1 and not in_question:
            idx_to_delete.append(curr_idx)

        # if both are in cell and before is after end, delete intervening lines
        elif lines_before_begin != -1 and lines_after_end != -1 and lines_after_end < lines_before_begin:
            del source[lines_after_end+1:lines_before_begin]
        
        else:
            # if there is an end comment, delete lines after that
            if lines_after_end != -1:
                del source[lines_after_end+1:]

            # if there is a begin comment, delete lines before that
            if lines_before_begin != -1:
                del source[:lines_before_begin]

        # update source
        cell["source"] = "\n".join(source)
    
    # reverse indices list so that we do not need to decrement while deleting
    idx_to_delete.reverse()

    # delete at indices
    for idx in idx_to_delete:
        del cells[idx]

    return notebook
